/*
 *   Emu48.h
 *
 *   This file is part of Emu48
 *
 *   Copyright (C) 1995 Sebastien Carlier
 *
 */

#include "types.h"

#ifdef WINCE

#include "wince.h"

#define wcstrncmp(a, b, c)						(wcsncmp(a, b, c))
#define _ASSERT(expr) ((void)0)

HWND hwndOwner;			// For the timers...

#else // WINCE
#define wcstrncmp(a, b, c)						(strncmp(a, b, c))

#define CreateFileCE	CreateFile
#define FindFirstFileCE FindFirstFile
#define CreateFileForMappingCE CreateFile
#endif // WINCE

// Emu48.c
extern int	   nRotation;
extern VOID    rotatePoint(DWORD *px, DWORD *py);
extern TCHAR szLicence[1024];
extern HPALETTE hPalette;
extern LPTSTR   szTitle;
extern HINSTANCE  hApp;
extern HWND    hWnd;
extern HDC     hWindowDC;
extern BOOL    bPort2IsShared;
extern BOOL    bAlwaysDisplayLog;
extern UINT    uTimer1Period;
extern VOID    SetWindowTitle(LPTSTR szString);
extern VOID    UpdateWindowStatus();

// Display.c
extern UINT    nBackgroundX;
extern UINT    nBackgroundY;
extern UINT    nBackgroundW;
extern UINT    nBackgroundH;
extern UINT    nLcdX;
extern UINT    nLcdY;
extern BOOL    bLcdDoubled;
extern LPBYTE  pbyLcd;
extern HDC     hLcdDC;
extern HDC     hMainDC;
extern BOOL BitBltRot(HDC hdcDest, int nXDest, int nYDest, int nWidth, int nHeight, 
					  HDC hdcSource, int nXSrc, int nYSrc, DWORD dwRop, int nRot);
extern VOID UpdateContrast(BYTE byContrast);
extern VOID SetLcdColor(UINT nId, UINT nRed, UINT nGreen, UINT nBlue);
extern VOID CreateLcdBitmap();
extern VOID DestroyLcdBitmap();
extern BOOL CreateMainBitmap(LPTSTR szFilename);
extern VOID DestroyMainBitmap();
extern VOID UpdateDisplayPointers();
extern VOID UpdateMainDisplay();
extern VOID UpdateMenuDisplay();
extern VOID WriteToMainDisplay(LPBYTE a, DWORD d, UINT s);
extern VOID WriteToMenuDisplay(LPBYTE a, DWORD d, UINT s);
extern VOID UpdateAnnunciators();
extern VOID ResizeWindow();

// Engine.c
extern BOOL     bInterrupt;
extern UINT     nState;
extern UINT     nNextState;
extern HANDLE   hThread;
extern CHIPSET  Chipset;
extern UINT     SwitchToState(UINT nNewState);
extern UINT     WorkerThread(LPVOID pParam);
extern DWORD    Npack(BYTE *a, UINT s);
extern VOID     Nunpack(BYTE *a, DWORD b, UINT s);

// Files.c
extern TCHAR    szEmu48Directory[260];
extern TCHAR    szCurrentDirectory[260];
extern TCHAR    szCurrentKml[260];
extern TCHAR    szBackupKml[260];
extern TCHAR    szCurrentFilename[260];
extern TCHAR    szBackupFilename[260];
extern TCHAR    szBufferFilename[260];
extern TCHAR    szPort2Filename[260];
extern LPBYTE  pbyRom;
extern DWORD   dwRomSize;
extern TCHAR    cCurrentRomType;
extern LPBYTE  pbyPort2;
extern BOOL    bPort2Writeable;
extern BOOL    bPort2IsShared;
extern DWORD   dwPort2Mask;
extern UINT    nPort2Bank;
extern BOOL    bBackup;
extern BOOL    MapRom(LPCTSTR szFilename);
extern VOID    UnmapRom();
extern BOOL    MapPort2(LPCTSTR szFilename);
extern VOID    UnmapPort2();
extern BOOL    PatchRom(LPCTSTR szFilename);
extern VOID    ResetDocument();
extern BOOL    NewDocument();
extern BOOL    OpenDocument(LPCTSTR szFilename);
extern BOOL    SaveDocument();
extern BOOL    SaveDocumentAs(LPCTSTR szFilename);
extern BOOL    SaveBackup();
extern BOOL    RestoreBackup();
extern BOOL    ResetBackup();
extern BOOL    GetOpenFilename();
extern BOOL    GetSaveAsFilename();
extern BOOL    GetLoadObjectFilename();
extern BOOL    GetSaveObjectFilename();
extern BOOL    LoadObject(LPCTSTR szFilename);
extern BOOL    SaveObject(LPCTSTR szFilename);
extern HBITMAP LoadBitmapFile(LPCTSTR szFilename);

// Timer.c
extern BOOL  bAccurateTimer;
extern UINT  uT1Period;
extern VOID  StartTimers();
extern VOID  StopTimers();
extern DWORD ReadT2();
extern VOID  SetT2(DWORD dwValue);
extern BYTE  ReadT1();
extern VOID  SetT1(BYTE byValue);

// MOps.c
extern LPBYTE  RMap[256];
extern LPBYTE  WMap[256];
extern VOID    Map(BYTE a, BYTE b);
extern VOID    Config();
extern VOID    Uncnfg();
extern VOID    Reset();
extern VOID    C_Eq_Id();
extern VOID    Npeek(BYTE *a, DWORD d, UINT s);
extern VOID    Nread(BYTE *a, DWORD d, UINT s);
extern VOID    Nwrite(BYTE *a, DWORD d, UINT s);
extern VOID    Nread2(BYTE *a, DWORD d);
extern VOID    Nwrite2(BYTE *a, DWORD d);
extern VOID    Nread5(BYTE *a, DWORD d);
extern VOID    Nwrite5(BYTE *a, DWORD d);
extern BYTE    Read2(DWORD d);
extern DWORD   Read5(DWORD d);
extern VOID    Write5(DWORD d, DWORD n);
extern VOID    ReadIO(BYTE *a, DWORD b, DWORD s);
extern VOID    WriteIO(BYTE *a, DWORD b, DWORD s);

// Keyboard.c
extern WORD Keyboard_GetIR();
extern VOID ScanKeyboard();
extern VOID KeyboardEvent(BOOL bPress, UINT out, UINT in);

// RPL.c
extern DWORD   RPL_SkipOb(DWORD d);
extern DWORD   RPL_ObjectSize(BYTE *o);
extern DWORD   RPL_CreateTemp(DWORD l);
extern DWORD   RPL_Pick(UINT l);
extern VOID    RPL_Replace(DWORD n);
extern VOID    RPL_Push(DWORD n);

// in External.c
extern VOID External(CHIPSET* w);

// Message Boxes
static __inline int InfoMessage(LPCTSTR szMessage)  {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_OK|MB_ICONINFORMATION|MB_SETFOREGROUND);}
static __inline int AbortMessage(LPCTSTR szMessage) {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_OK|MB_ICONSTOP|MB_SETFOREGROUND);}
static __inline int YesNoMessage(LPCTSTR szMessage) {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_YESNO|MB_ICONEXCLAMATION|MB_SETFOREGROUND);}
static __inline int YesNoCancelMessage(LPCTSTR szMessage) {return MessageBox(hWnd, szMessage, szTitle, MB_APPLMODAL|MB_YESNOCANCEL|MB_ICONEXCLAMATION|MB_SETFOREGROUND);}

#ifdef WINCE
void CopyMemory (PVOID bpDest, CONST VOID *bpSource, DWORD dwLength);
void FillMemory (PVOID bpDest, DWORD dwLength, BYTE byFill);
#endif

// Missing Win32 API calls
static __inline LPTSTR   DuplicateString(LPCTSTR szString)
{
	UINT  uLength =  _tcslen(szString) + 1;
	LPTSTR szDup   = LocalAlloc(0,(uLength+1) * sizeof(TCHAR));
	if (IsBadWritePtr(szDup, (uLength+1)*sizeof(TCHAR)))
		return NULL;

	lstrcpy(szDup, szString);
	//CopyMemory(szDup,szString,uLength);
	return szDup;
}
