#include "pch.h"
#include "types.h"
#include "resource.h"

/* Funciones no implementadas de WIN32 y Timers */


/* Timers */

#ifdef WINCE

void vprintf( LPCTSTR lpsData );
MMRESULT timeKillEvent( UINT uTimerID );

#define MAX_TIMERS		16

static struct timer_info
{
	UINT			hTimer;
	LPTIMECALLBACK	lpTimeProc;
	DWORD			dwUser;
	UINT			fuEvent;
	UINT			uDelay;
	HANDLE			hTimerThread;
	DWORD			idTimerThread;
};


MMRESULT timeBeginPeriod( UINT uPeriod )
{
	return TIMERR_NOERROR;
}

MMRESULT timeEndPeriod( UINT uPeriod )
{
	return TIMERR_NOERROR;
}


static DWORD WINAPI	MMTimerThread( struct timer_info FAR  * timer_data)
{
	MSG	msg;

	timer_data->hTimer=SetTimer( NULL, 0, timer_data->uDelay, (TIMERPROC) NULL);

	while (GetMessage(&msg, NULL, 0, 0))
	{
		if(msg.message==WM_TIMER)
		{	
			(*timer_data->lpTimeProc)(timer_data->idTimerThread,msg.message,timer_data->dwUser,0,0);
			if(timer_data->fuEvent==TIME_ONESHOT)
				timeKillEvent(timer_data->idTimerThread);
		}
		DispatchMessage(&msg);
	}
	KillTimer(NULL,timer_data->hTimer);
	LocalFree(timer_data);
	ExitThread(0);
	return 0;
}


MMRESULT timeSetEvent( UINT uDelay, UINT uResolution,
					   LPTIMECALLBACK lpTimeProc, DWORD dwUser,
					   UINT fuEvent )
{
	struct timer_info FAR	 *timer_data;

	timer_data=(struct timer_info FAR *)LocalAlloc(LPTR,sizeof(struct timer_info));
	timer_data->lpTimeProc=lpTimeProc;
	timer_data->dwUser=dwUser;
	timer_data->fuEvent=fuEvent;
	timer_data->uDelay=uDelay;
	timer_data->hTimerThread=CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)&MMTimerThread,timer_data,0,&(timer_data->idTimerThread));
	if(!timer_data->hTimerThread)
		return 0;
	CloseHandle(timer_data->hTimerThread);
	return (UINT) timer_data->idTimerThread;
}


MMRESULT timeKillEvent( UINT uTimerID )
{
	return PostThreadMessage(uTimerID,WM_QUIT,0,0) ? TIMERR_NOERROR : MMSYSERR_INVALPARAM;
}



/* WIN32 Missing Functions */

// Directory Functions

static TCHAR current_directory[256]=_T("\\");

HANDLE CreateFileCE(LPCTSTR lpFileName, DWORD dwDesiredAccess, DWORD dwShareMode, LPSECURITY_ATTRIBUTES lpSecurityAttributes, DWORD dwCreationDistribution, DWORD dwFlagsAndAttributes, HANDLE hTemplateFile)
{
	TCHAR	buffer[260];
	HANDLE	aux;

	if(lpFileName[0]!=__T('\\'))
	{
		lstrcpy(buffer,current_directory);
		lstrcat(buffer,lpFileName);
		return CreateFile(buffer, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDistribution, dwFlagsAndAttributes, hTemplateFile );
	}
	return CreateFile(lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDistribution, dwFlagsAndAttributes, hTemplateFile );
}

HANDLE CreateFileForMappingCE(LPCTSTR lpFileName, DWORD dwDesiredAccess, DWORD dwShareMode, LPSECURITY_ATTRIBUTES lpSecurityAttributes, DWORD dwCreationDistribution, DWORD dwFlagsAndAttributes, HANDLE hTemplateFile)
{
	TCHAR	buffer[260];
	HANDLE	aux;
	TCHAR	bufaux[80];

	if(lpFileName[0]!=__T('\\'))
	{
		lstrcpy(buffer,current_directory);
		lstrcat(buffer,lpFileName);
//		wsprintf(bufaux,_T("OpenMap: %s"),buffer);
//		vprintf(bufaux);
		aux=CreateFileForMapping(buffer, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDistribution, dwFlagsAndAttributes, hTemplateFile );
		if (aux==INVALID_HANDLE_VALUE)
		{
			wsprintf(bufaux,_T("Open failed for file; Error: %u"),
				GetLastError());
//			vprintf(bufaux);
		}
		return aux;
	}
	return CreateFileForMapping(lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDistribution, dwFlagsAndAttributes, hTemplateFile );
}

HANDLE FindFirstFileCE(LPCTSTR lpFileName, LPWIN32_FIND_DATA lpFindFileData)
{
	HANDLE	i;
	TCHAR	buffer[260];

	if(lpFileName[0]!=__T('\\'))
	{
		lstrcpy(buffer,current_directory);
		lstrcat(buffer,lpFileName);
		return FindFirstFile(buffer, lpFindFileData);
	}
	return FindFirstFile( lpFileName, lpFindFileData);
}

DWORD GetCurrentDirectory( 

	DWORD nBufferLength, // size, in characters, of directory buffer 

	LPTSTR lpBuffer // address of buffer for current directory 

)
{

	UINT i;

	for(i=0;current_directory[i] && ( i < nBufferLength );i++)
		lpBuffer[i]=current_directory[i];
	lpBuffer[i]=0;

	return (DWORD) i;

}

BOOL SetCurrentDirectory( LPCTSTR lpPathName )
{
	lstrcpy( current_directory, lpPathName);
	return	TRUE;
}

// Graphic Functions 

#define MAX_HDC		8

static 
struct hdc_pos
{
	HDC		hdc;
	HWND	hwnd;
	POINT	pCurrentPos;
}
	hdc_info[MAX_HDC];

void Initialize_GraphFun( void )
{
	int	i;
	for(i=0;i<MAX_HDC;i++)
		hdc_info[i].hdc= (HDC) NULL;
}


HDC BeginPaintCE (HWND hwnd, LPPAINTSTRUCT lpPaint)
{
	int i;

	for(i=0;i<MAX_HDC;i++)
		if(hdc_info[i].hdc==(HDC) NULL) break;
	if(i==MAX_HDC) return NULL;
	hdc_info[i].pCurrentPos.x=0;
	hdc_info[i].pCurrentPos.y=0;
	hdc_info[i].hwnd=hwnd;
	return( hdc_info[i].hdc=BeginPaint(hwnd, lpPaint));
}

BOOL EndPaintCE(HWND hWnd, LPPAINTSTRUCT lpPaint)
{
	int i;

	for(i=0;i<MAX_HDC;i++)
		if(hdc_info[i].hdc==lpPaint->hdc) break;
	if(i==MAX_HDC) return EndPaint(hWnd,lpPaint); //FALSE
	if(EndPaint(hWnd,lpPaint))
	{
		hdc_info[i].hdc= (HDC) NULL;
		return TRUE;
	}
	return FALSE;
}

HDC GetDC_CE(HWND hWnd)
{
	int i;

	for(i=0;i<MAX_HDC;i++)
		if(hdc_info[i].hdc==(HDC) NULL) break;
	if(i==MAX_HDC) return NULL;
	hdc_info[i].pCurrentPos.x=0;
	hdc_info[i].pCurrentPos.y=0;
	hdc_info[i].hwnd=hWnd;
	return( hdc_info[i].hdc=GetDC(hWnd));
}

int ReleaseDC_CE(HWND hWnd, HDC hDC)
{
	int i;

	for(i=0;i<MAX_HDC;i++)
		if( (hdc_info[i].hdc==hDC) ) break;
	if(i==MAX_HDC) return ReleaseDC(hWnd, hDC); //FALSE
	hdc_info[i].hdc= (HDC) NULL;
	return ReleaseDC(hWnd, hDC);
}
	

BOOL LineTo( HDC hdc, int nXEnd, int nYEnd )
{
	int i;
	POINT	pLine[2];

	for(i=0;i<MAX_HDC;i++)
		if(hdc_info[i].hdc==hdc) break;
	if(i==MAX_HDC) return FALSE;
	pLine[0]=hdc_info[i].pCurrentPos;
	pLine[1].x=nXEnd;
	pLine[1].y=nYEnd;
	if (Polyline(hdc, pLine, 2))
	{
		hdc_info[i].pCurrentPos=pLine[1];
		return TRUE;
	}
	return	FALSE;
}

BOOL MoveToEx( HDC hdc, int X, int Y, LPPOINT lppoint)
{
	int i;

	for(i=0;i<MAX_HDC;i++)
		if(hdc_info[i].hdc==hdc) break;
	if(i==MAX_HDC) return FALSE;
	hdc_info[i].pCurrentPos.x=X;
	hdc_info[i].pCurrentPos.y=Y;
	return TRUE;
}


// Esto es para debugging

TCHAR	buffer[80];
extern HINSTANCE hApp;
extern HWND      hWnd;


static BOOL CALLBACK ShowPrintf(HWND hDlg, UINT message, DWORD wParam, LONG lParam)
{
	
	switch (message)
	{
	case WM_INITDIALOG:
		SetDlgItemText(hDlg,IDC_PTEXT,buffer);
		return TRUE;
		
	case WM_COMMAND:
		if (wParam == IDOK)
			EndDialog(hDlg, IDOK);
		break;		
	}
	return FALSE;
    UNREFERENCED_PARAMETER(lParam);
    UNREFERENCED_PARAMETER(hDlg);
}

void vprintf( LPCTSTR lpsData )
{
	lstrcpy(buffer,lpsData);
	DialogBox(hApp,MAKEINTRESOURCE( IDD_PRINTF ),hWnd, (DLGPROC) ShowPrintf);
}


// This is for solving SH3 word alignment problem
#ifdef SH3

typedef struct _tagBITMAPINFOHEADER {
        UNALIGNED DWORD      biSize;
        UNALIGNED LONG       biWidth;
        UNALIGNED LONG       biHeight;
		UNALIGNED WORD       biPlanes;
        UNALIGNED WORD       biBitCount;
        UNALIGNED DWORD      biCompression;
        UNALIGNED DWORD      biSizeImage;
        UNALIGNED LONG       biXPelsPerMeter;
        UNALIGNED LONG       biYPelsPerMeter;
        UNALIGNED DWORD      biClrUsed;
        UNALIGNED DWORD      biClrImportant;
} UNALIGNED BITMAPINFOHEADER_SH3, FAR *LPBITMAPINFOHEADER_SH3, *PBITMAPINFOHEADER_SH3;

typedef struct _tagBITMAPINFO {
    BITMAPINFOHEADER    bmiHeader;
    RGBQUAD             bmiColors[1];
} UNALIGNED BITMAPINFO_SH3,FAR *LPBITMAPINFO_SH3, *PBITMAPINFO_SH3;

HBITMAP CreateDIBSectionSH3(HDC hdc, const BITMAPINFO_SH3 *pbmi, UINT iUsage, void *ppvBits, HANDLE hSection, DWORD dwOffset)
{
	BITMAPINFO	*pbmi_copy;
	UINT		i;
	HBITMAP		retvalue;
	
	pbmi_copy=(BITMAPINFO *)LocalAlloc(LPTR, sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD));
	
	pbmi_copy->bmiHeader=pbmi->bmiHeader;
	for(i=0;i<256;i++)
		pbmi_copy->bmiColors[i]=pbmi->bmiColors[i];
	retvalue= CreateDIBSection(hdc,pbmi_copy,iUsage,ppvBits,hSection,dwOffset);
	LocalFree(pbmi_copy);
	return retvalue;
}


#endif //SH3

#endif
 